/**
 * @file main.cpp
 * @date 2017-05-28
 */

 // Copyright (C) 2016-2017 Usagi
 // Licensed under the MIT License.

 // for Siv3D August 2016 v2


#include <Siv3D.hpp>

#include "HamFramework.hpp"
#include "tinyxml2.h"

#include "mqoparser.hpp"

// XCh̃TCY(Ch摜)
//#define SL_W (12192000)
//#define SL_H  (6858000)

// XCh̃TCY(ʂɂ킹(16:9))
#define SL_W (9144000)
#define SL_H (5143500)

// 1mm̑傫
#define UNIT (36000)
// 1/72C`1pt̑傫
#define UNITPT (12700)
// 0.25 pt ̑傫
#define UNITPTQ (3175)

#define FSL L"Large"
#define FSM L"Medium"
#define FSS L"Small"

#define SL_AMIN L"slider_alphamin"
#define SL_FOVD L"slider_fovd"
#define SL_THICKQ L"slider_thickq"

#define PX_W (960)
#define PX_H (540)

// LɂĂ
#define MAXDOTSIDE (64)

/// <summary>1hbg̋`</summary>
struct DotBox {
	/// <summary>ƕ</summary>
	Rect rc;

	/// <summary>F</summary>
	Color col;
};

/// <summary>FL2D1</summary>
struct LineLine {
	/// <summary>n_ƏI_</summary>
	Line line;

	/// <summary>F</summary>
	Color col;

	/// <summary>傫ւ̋`</summary>
	Rect rc;

	/// <summary>falseƏ傫</summary>
	bool flipH;
	/// <summary>falseƏ傫</summary>
	bool flipV;

	/// <summary>`ƃtbvtO肷</summary>
	void calcRect() {
		auto x0 = line.begin.x;
		auto y0 = line.begin.y;
		auto x1 = line.end.x;
		auto y1 = line.end.y;

		this->flipH = (x0 > x1);
		this->flipV = (y0 > y1);

		if (this->flipH) {
			auto tmp = x0;
			x0 = x1;
			x1 = tmp;
		}
		if (this->flipV) {
			auto tmp = y0;
			y0 = y1;
			y1 = tmp;
		}

		this->rc.set({ x0,y0 }, { x1 - x0, y1 - y0 });
	};
};


/// <summary>`p</summary>
struct ColorLine3Ds {
	/// <summary>3Di[</summary>
	Array<Line3D> lines;
	/// <summary>̐F</summary>
	Color col;
};

/// <summary>3x3sBOpenCVJIvWFNVɂ</summary>
template <class T>
class Mat3x3 {
public:
	/// <summary>vfBX^Ŋi[</summary>
	T e[9];

	/// <summary>RXgN^B0s</summary>
	Mat3x3() {
		for (int i = 0; i < 9; ++i) {
			this->e[i] = 0.0;
		}
	};

	/// <summary>Y𐂒Ƃ ang ZX]s</summary>
	void setHead(T ang) {
		auto c = cos(ang);
		auto s = sin(ang);

		this->e[0] = c;
		this->e[1] = 0;
		this->e[2] = -s;

		this->e[3] = 0;
		this->e[4] = 1;
		this->e[5] = 0;

		this->e[6] = s;
		this->e[7] = 0;
		this->e[8] = c;
	};
	void setPitch(T ang) {
		auto c = cos(ang);
		auto s = sin(ang);

		this->e[0] = 1;
		this->e[1] = 0;
		this->e[2] = 0;

		this->e[3] = 0;
		this->e[4] = c;
		this->e[5] = s;

		this->e[6] = 0;
		this->e[7] = -s;
		this->e[8] = c;
	};

	/// <summary>1s33x3s̐ςV</summary>
	/// <param name="v">炩xNg</param>
	Vec3 leftMuled(const Vec3& v) const {
		Vec3 ret;
		ret.x = v.x * this->e[0] + v.y * this->e[3] + v.z * this->e[6];
		ret.y = v.x * this->e[1] + v.y * this->e[4] + v.z * this->e[7];
		ret.z = v.x * this->e[2] + v.y * this->e[5] + v.z * this->e[8];
		return ret;
	};

	/// <summary>3x3s납|</summary>
	static Mat3x3 mulMat(const Mat3x3& l, const Mat3x3& r) {
		Mat3x3 ret;
		for (int i = 0; i < 3; ++i) {
			for (int j = 0; j < 3; ++j) {
				T sum = 0.0;
				for (int k = 0; k < 3; ++k) {
					sum += l.e[i * 3 + k] * r.e[k * 3 + j];
				}
				ret.e[i * 3 + j] = sum;
			}
		}
		return ret;
	};
};

/// <summary>܂Ƃߍ\</summary>
struct Data {
	/// <summary>GUI</summary>
	Array<GUI> guis;

	/// <summary>`p</summary>
	Array<ColorLine3Ds> colorsets;


	//// J
	/// <summary>}EXʒu</summary>
	Optional<Point> prePos;

	/// <summary>J</summary>
	Vec3 basePYR = { 0.5236, -s3d::Math::HalfPi + 0.5236,0 };
	/// <summary>ZoJ</summary>
	Vec3 curPYR = { 0.5236, -s3d::Math::HalfPi + 0.5236,0 };
	/// <summary>J̋</summary>
	double camDist = 1000.0;
	/// <summary>JĂ3D_</summary>
	Vec3 camTarget = { 0,0,0 };
};


/// <summary>Z𔽓]VVec3Ԃ</summary>
Vec3 zReved(const Vec3& v) {
	return Vec3(v.x, v.y, -v.z);
}

/// <summary>ft@C̑f𐶐</summary>
void makeModel(tinyxml2::XMLDocument& doc) {

	{
		doc.InsertFirstChild(doc.NewDeclaration());

		auto e1 = doc.NewElement("model");
		e1->SetAttribute("unit", "millimeter");
		e1->SetAttribute("xml:lang", "en-US");
		e1->SetAttribute("xmlns:m", "http://schemas.microsoft.com/3dmanufacturing/material/2015/02");
		e1->SetAttribute("xmlns", "http://schemas.microsoft.com/3dmanufacturing/core/2015/02");

		{
			auto e2 = doc.NewElement("resources");
			{
				auto e3 = doc.NewElement("basematerials");
				e3->SetAttribute("id", "1");
				auto e4 = doc.NewElement("base");
				e4->SetAttribute("name", "Forest Green");
				e4->SetAttribute("displaycolor", "#6DB555FF");
				e3->InsertEndChild(e4);
				e2->InsertEndChild(e3);
			}

			{
				auto e3 = doc.NewElement("object");
				e3->SetAttribute("id", "2");
				e3->SetAttribute("type", "model");

				auto e4 = doc.NewElement("mesh");
				auto e5 = doc.NewElement("vertices");
				for (int i = 0; i < 4; ++i) {
					auto e6 = doc.NewElement("vertex");
					e6->SetAttribute("x", Format(PyFmt, L"{}", (i&1)*100).narrow().c_str());
					e6->SetAttribute("y", Format(PyFmt, L"{}", (i/2)*100).narrow().c_str());
					e6->SetAttribute("z", "-300");
					e5->InsertEndChild(e6);
				}
				e4->InsertEndChild(e5);

				auto e7 = doc.NewElement("triangles");
				for (int i = 0; i < 1; ++i) {
					auto e8 = doc.NewElement("triangle");
					e8->SetAttribute("v1", "0");
					e8->SetAttribute("v2", "1");
					e8->SetAttribute("v3", "2");
					e8->SetAttribute("pid", "1");
					e8->SetAttribute("p1", "0");

					e7->InsertEndChild(e8);
				}
				e4->InsertEndChild(e7);

				e3->InsertEndChild(e4);
				e2->InsertEndChild(e3);
			}

			{
				auto e3 = doc.NewElement("object");
				e3->SetAttribute("id", "3");
				e3->SetAttribute("type", "model");
				auto e4 = doc.NewElement("components");
				auto e5 = doc.NewElement("component");
				e5->SetAttribute("objectid", "2");
				e4->InsertEndChild(e5);
				e3->InsertEndChild(e4);
				e2->InsertEndChild(e3);
			}

			e1->InsertEndChild(e2);
		}
		{
			auto e2 = doc.NewElement("build");
			auto e3 = doc.NewElement("item");
			e3->SetAttribute("objectid", "3");
			e3->SetAttribute("transform", "1 0 0 0 1 0 0 0 1 0 0 0");
			e2->InsertEndChild(e3);
			e1->InsertEndChild(e2);
		}
		doc.InsertEndChild(e1);
	}
}


/// <summary>nvSpPr ̕</summary>
tinyxml2::XMLElement* makeName(tinyxml2::XMLDocument& doc, int id) {
	auto e4 = doc.NewElement("p:nvSpPr");
	{
		char s[_MAX_PATH];
		sprintf_s(s, _MAX_PATH, "%d", id);

		auto e5 = doc.NewElement("p:cNvPr");
		e5->SetAttribute("id", s);
		e5->SetAttribute("name", "");
		e4->InsertEndChild(e5);
	}
	{
		auto e5 = doc.NewElement("p:cNvSpPr");
		e4->InsertEndChild(e5);
	}
	{
		auto e5 = doc.NewElement("p:nvPr");
		e4->InsertEndChild(e5);
	}
	return e4;
}

/// <summary>nvGrpSpPr ̕</summary>
tinyxml2::XMLElement* makeNameGroup(tinyxml2::XMLDocument& doc, int id) {
	auto e4 = doc.NewElement("p:nvGrpSpPr");
	{
		char s[_MAX_PATH];
		sprintf_s(s, _MAX_PATH, "%d", id);

		auto e5 = doc.NewElement("p:cNvPr");
		e5->SetAttribute("id", s);
		e5->SetAttribute("name", "");
		e4->InsertEndChild(e5);
	}
	{
		auto e5 = doc.NewElement("p:cNvGrpSpPr");
		e4->InsertEndChild(e5);
	}
	{
		auto e5 = doc.NewElement("p:nvPr");
		e4->InsertEndChild(e5);
	}
	return e4;
}

/// <summary>nvCxnSpPr ̕</summary>
tinyxml2::XMLElement* makeNameCxn(tinyxml2::XMLDocument& doc, int id) {
	auto e4 = doc.NewElement("p:nvCxnSpPr");
	{
		char s[_MAX_PATH];
		sprintf_s(s, _MAX_PATH, "%d", id);

		auto e5 = doc.NewElement("p:cNvPr");
		e5->SetAttribute("id", s);
		e5->SetAttribute("name", "");
		e4->InsertEndChild(e5);
	}
	{
		auto e5 = doc.NewElement("p:cNvCxnSpPr");
		e4->InsertEndChild(e5);
	}
	{
		auto e5 = doc.NewElement("p:nvPr");
		e4->InsertEndChild(e5);
	}
	return e4;
}


/// <summary>a:xfrm Gg𐶐</summary>
/// <param name="isCh">chOffsetȂǂǉ邩ǂ</param>
tinyxml2::XMLElement* makeXform(tinyxml2::XMLDocument& doc, Rect rc, bool isCh,
		bool isH = false, bool isV = false) {
	char x[_MAX_PATH];
	char y[_MAX_PATH];
	char cx[_MAX_PATH];
	char cy[_MAX_PATH];

	sprintf_s(x, _MAX_PATH, "%d", rc.x);
	sprintf_s(y, _MAX_PATH, "%d", rc.y);
	sprintf_s(cx, _MAX_PATH, "%d", rc.w);
	sprintf_s(cy, _MAX_PATH, "%d", rc.h);

	auto e5 = doc.NewElement("a:xfrm");
	if (isH) {
		e5->SetAttribute("flipH", 1);
	}
	if (isV) {
		e5->SetAttribute("flipV", 1);
	}

	{
		auto e6 = doc.NewElement("a:off");
		e6->SetAttribute("x", x);
		e6->SetAttribute("y", y);
		e5->InsertEndChild(e6);
	}
	{
		auto e6 = doc.NewElement("a:ext");
		e6->SetAttribute("cx", cx);
		e6->SetAttribute("cy", cy);
		e5->InsertEndChild(e6);
	}
	if (isCh) {
		{
			auto e6 = doc.NewElement("a:chOff");
			e6->SetAttribute("x", x);
			e6->SetAttribute("y", y);
			e5->InsertEndChild(e6);
		}
		{
			auto e6 = doc.NewElement("a:chExt");
			e6->SetAttribute("cx", cx);
			e6->SetAttribute("cy", cy);
			e5->InsertEndChild(e6);
		}
	}

	return e5;
}

/// <summary>.xml𐶐</summary>
/// <param name="dbs">hbgz</param>
/// <param name="lls">Cz</param>
/// <param name="thick">̒Pʂőw肷</param>
std::string makeXmlString(const Array<DotBox>& dbs, const Array<LineLine>& lls,
		int thick) {
	LOG_DEBUG(L"makeXmlString");

	tinyxml2::XMLDocument doc;
	doc.InsertFirstChild(doc.NewDeclaration());

	auto e1 = doc.NewElement("p:sld");
	e1->SetAttribute("xmlns:a", "http://schemas.openxmlformats.org/drawingml/2006/main");
	e1->SetAttribute("xmlns:r", "http://schemas.openxmlformats.org/officeDocument/2006/relationships");
	e1->SetAttribute("xmlns:p", "http://schemas.openxmlformats.org/presentationml/2006/main");

	int index = 1;

	{
		auto e2 = doc.NewElement("p:cSld");
		{
			auto e3 = doc.NewElement("p:spTree");
			{
				auto e4 = makeNameGroup(doc, index);
				++index;
				e3->InsertEndChild(e4);
			}
			{
				auto e4 = doc.NewElement("p:grpSpPr");
				auto e5 = makeXform(doc, Rect(0, 0, 0, 0), true);
				e4->InsertEndChild(e5);
				e3->InsertEndChild(e4);
			}

			if (dbs.size() > 0) { // ̕
				auto e4 = doc.NewElement("p:grpSp");
				{
					auto e5 = makeNameGroup(doc, index);
					++index;
					e4->InsertEndChild(e5);
				}
				{
					auto e5 = doc.NewElement("p:grpSpPr");
					auto e6 = makeXform(doc, Rect(0,0, SL_W,SL_H), true);
					e5->InsertEndChild(e6);
					e4->InsertEndChild(e5);
				}
				for (const auto& db : dbs) {
					auto e5 = doc.NewElement("p:sp");
					{
						auto e6 = makeName(doc, index);
						++index;
						e5->InsertEndChild(e6);
					}
					{
						auto e6 = doc.NewElement("p:spPr");
						{
							auto e7 = makeXform(doc, db.rc, false);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:prstGeom");
							e7->SetAttribute("prst", "rect");
							auto e8 = doc.NewElement("a:avLst");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:solidFill");
							auto e8 = doc.NewElement("a:srgbClr");
							char col[_MAX_PATH];
							sprintf_s(col, _MAX_PATH, "%02X%02X%02X", db.col.r, db.col.g, db.col.b);
							e8->SetAttribute("val", col);
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:ln");
							auto e8 = doc.NewElement("a:noFill");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						e5->InsertEndChild(e6);
					}
					{
						auto e6 = doc.NewElement("p:style");
						{
							auto e7 = doc.NewElement("a:lnRef");
							e7->SetAttribute("idx", "1");
							auto e8 = doc.NewElement("a:schemeClr");
							e8->SetAttribute("val", "accent1");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:fillRef");
							e7->SetAttribute("idx", "1");
							auto e8 = doc.NewElement("a:schemeClr");
							e8->SetAttribute("val", "accent1");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:effectRef");
							e7->SetAttribute("idx", "0");
							auto e8 = doc.NewElement("a:schemeClr");
							e8->SetAttribute("val", "accent1");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:fontRef");
							e7->SetAttribute("idx", "minor");
							auto e8 = doc.NewElement("a:schemeClr");
							e8->SetAttribute("val", "lt1");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						e5->InsertEndChild(e6);
					}
					e4->InsertEndChild(e5);
				}
				e3->InsertEndChild(e4);
			}

			if (lls.size() > 0) { // ̕
				auto e4 = doc.NewElement("p:grpSp");
				{
					auto e5 = makeNameGroup(doc, index);
					++index;
					e4->InsertEndChild(e5);
				}
				{
					auto e5 = doc.NewElement("p:grpSpPr");
					auto e6 = makeXform(doc, Rect(0, 0, SL_W, SL_H), true);
					e5->InsertEndChild(e6);
					e4->InsertEndChild(e5);
				}
				for (const auto& ll : lls) {
					auto e5 = doc.NewElement("p:cxnSp");
					e4->InsertEndChild(e5);

					{
						auto e6 = makeNameCxn(doc, index);
						++index;
						e5->InsertEndChild(e6);
					}
					{
						auto e6 = doc.NewElement("p:spPr");
						{
							auto e7 = makeXform(doc, ll.rc, false, ll.flipH, ll.flipV);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:prstGeom");
							e7->SetAttribute("prst", "line");
							auto e8 = doc.NewElement("a:avLst");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						{ // 2Ggǉ
							auto e7 = doc.NewElement("a:ln");
							e7->SetAttribute("cap", "rnd");
							e7->SetAttribute("w", std::to_string(thick).c_str());
							{
								auto e8 = doc.NewElement("a:solidFill");
								auto e9 = doc.NewElement("a:srgbClr");
								char col[_MAX_PATH];
								sprintf_s(col, _MAX_PATH, "%02X%02X%02X", ll.col.r, ll.col.g, ll.col.b);
								e9->SetAttribute("val", col);
								e8->InsertEndChild(e9);
								e7->InsertEndChild(e8);
							}
							e6->InsertEndChild(e7);
						}
						e5->InsertEndChild(e6);
					}
					{
						auto e6 = doc.NewElement("p:style");
						{
							auto e7 = doc.NewElement("a:lnRef");
							e7->SetAttribute("idx", "1");
							auto e8 = doc.NewElement("a:schemeClr");
							e8->SetAttribute("val", "accent1");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:fillRef");
							e7->SetAttribute("idx", "0");
							auto e8 = doc.NewElement("a:schemeClr");
							e8->SetAttribute("val", "accent1");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:effectRef");
							e7->SetAttribute("idx", "0");
							auto e8 = doc.NewElement("a:schemeClr");
							e8->SetAttribute("val", "accent1");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						{
							auto e7 = doc.NewElement("a:fontRef");
							e7->SetAttribute("idx", "minor");
							auto e8 = doc.NewElement("a:schemeClr");
							e8->SetAttribute("val", "tx1");
							e7->InsertEndChild(e8);
							e6->InsertEndChild(e7);
						}
						e5->InsertEndChild(e6);
					}
					e4->InsertEndChild(e5);
				}
				e3->InsertEndChild(e4);
			}

			e2->InsertEndChild(e3);
		}
		{
			auto e3 = doc.NewElement("p:extLst");
			auto e4 = doc.NewElement("p:ext");
			e4->SetAttribute("uri", "{BB962C8B-B14F-4D97-AF65-F5344CB8AC3E}");
			auto e5 = doc.NewElement("p14:creationId");
			e5->SetAttribute("xmlns:p14", "http://schemas.microsoft.com/office/powerpoint/2010/main");
			int id = s3d::Random<int>(1, 0x7FFFFFFF);
			e5->SetAttribute("val", id);
			e4->InsertEndChild(e5);
			e3->InsertEndChild(e4);
			e2->InsertEndChild(e3);
		}
		e1->InsertEndChild(e2);
	}

	{
		auto e2 = doc.NewElement("p:clrMapOvr");
		auto e3 = doc.NewElement("a:masterClrMapping");
		e2->InsertEndChild(e3);
		e1->InsertEndChild(e2);
	}
	doc.InsertEndChild(e1);

	tinyxml2::XMLPrinter printer;
	doc.Print(&printer);

	std::string ret = std::string(printer.CStr());

	LOG_DEBUG(L"makeXmlString leave");
	return ret;
}


/// <summary>ۑ</summary>
/// <param name="thickQ">̑ 1/4 pt </param>
bool savePptx(const Array<DotBox>& dbs, const Array<LineLine>& lls, const String& outPath,
		int thickQ) {
	bool result;

	String files[] = {
		L"[Content_Types].xml",

		L"_rels/.rels",

		L"docProps/app.xml",
		L"docProps/core.xml",
		L"docProps/thumbnail.jpeg",

		L"ppt/_rels/presentation.xml.rels",

		L"ppt/slideLayouts/_rels/slideLayout1.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout2.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout3.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout4.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout5.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout6.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout7.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout8.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout9.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout10.xml.rels",
		L"ppt/slideLayouts/_rels/slideLayout11.xml.rels",

		L"ppt/slideLayouts/slideLayout1.xml",
		L"ppt/slideLayouts/slideLayout2.xml",
		L"ppt/slideLayouts/slideLayout3.xml",
		L"ppt/slideLayouts/slideLayout4.xml",
		L"ppt/slideLayouts/slideLayout5.xml",
		L"ppt/slideLayouts/slideLayout6.xml",
		L"ppt/slideLayouts/slideLayout7.xml",
		L"ppt/slideLayouts/slideLayout8.xml",
		L"ppt/slideLayouts/slideLayout9.xml",
		L"ppt/slideLayouts/slideLayout10.xml",
		L"ppt/slideLayouts/slideLayout11.xml",

		L"ppt/slideMasters/_rels/slideMaster1.xml.rels",
		L"ppt/slideMasters/slideMaster1.xml",

		L"ppt/theme/theme1.xml",


		L"ppt/presentation.xml",
		L"ppt/presProps.xml",
		L"ppt/tableStyles.xml",
		L"ppt/viewProps.xml",

		L"ppt/slides/_rels/slide1.xml.rels",

		L"ppt/slides/slide1.xml",
	};


	ZIPWriter zw(outPath);

	for (int i = 0; i < 36; ++i) {
		auto local = Format(PyFmt, L"template/{}", files[i]);

		BinaryReader br(local);
		auto ba = br.readWhole();
		zw.addFile(files[i], ba.data(), ba.size());
	}

	{ // }`XCh
		std::string slide = makeXmlString(dbs, lls, thickQ * UNITPTQ);
		result = zw.addFile(L"ppt/slides/slide1.xml", slide.c_str(), slide.size());
		if (result == false) {
			zw.close();
			return false;
		}
	}

	zw.close();

	LOG_DEBUG(L"savePptx leave");
	return true;
}



using MyApp = ham::SceneManager<String, Data>;

struct MainScene : MyApp::Scene {
public:
	/// <summary></summary>
	void init() override {
		Window::SetBaseSize(PX_W, PX_H);
		Window::Resize(PX_W, PX_H);
		Window::SetTitle(L"mqo2pptx ver.0.3");

		FontAsset::Register(FSS, 16);

		Graphics::SetBackground({ 0xee,0xee,0xee });

		Graphics3D::SetAntiAliasing(AntiAliasing::Default);

		int uiSize = 24;

			{
				GUI gui(GUIStyle::Default);
				gui.setTitle(L"\ݒ/v");
				gui.addln(L"cbAxis", GUICheckBox::Create({ L"\" }, {}));

				gui.add(GUIHorizontalLine::Create());

				gui.add(L"txfps", GUIText::Create(L"-"));
				gui.add(GUIText::Create(L"fps"));
				gui.addln(L"tx1", GUIText::Create(L"-", uiSize));


				const Rect rc = gui.getRect();
				gui.setPos(4, Window::Height() - rc.h - 4);

				m_data->guis.push_back(gui);
			}

			{
				GUI gui(GUIStyle::Default);
				gui.setPos(4, 4);
				gui.setTitle(L"}`o͐ݒ");

				gui.addln(L"cbForceCol", GUICheckBox::Create({ L"FSčɂ" }, { 0 }));

				gui.add(GUIText::Create(L"FOV(x)"));
				gui.addln(L"btSet30", GUIButton::Create(L"30x"));
				gui.add(SL_FOVD, GUISlider::Create(5,90, 30, 172));
				gui.addln(SL_FOVD, GUIText::Create(L"30", uiSize));

				gui.add(GUIText::Create(L""));
				gui.add(SL_THICKQ, GUISlider::Create(1, 20, 4, 20*4));
				gui.addln(SL_THICKQ, GUIText::Create(L"1 pt", uiSize));

				gui.addln(L"btSave", GUIButton::Create(L"pptxo͂"));

				m_data->guis.push_back(gui);
			}

			{
				GUI gui(GUIStyle::Default);
				gui.setPos(320, 4);
				gui.setTitle(L"hbgGo͐ݒ");

				gui.addln(GUIText::Create(L"F̐ݒ"));
				gui.addln(L"rbBmpMode", GUIRadioButton::Create({
					L"Ȃ",L"",L"", L"" },
					{ 2 }));

				gui.add(GUIHorizontalLine::Create());

				gui.addln(GUIText::Create(L"Lsx̉"));
				gui.add(SL_AMIN, GUISlider::Create(0,255, 128, 128));
				gui.addln(SL_AMIN, GUIText::Create(L"128", uiSize));


				const Rect rc = gui.getRect();
				gui.setPos(Window::Width() - rc.w - 4, 4);
				m_data->guis.push_back(gui);
			}

	}

	/// <summary>tpX͂Ƃăt@Cۑ</summary>
	bool saveDots(const String& fileName) {
		auto outNames = Dialog::GetSave({{L"PowerPoint(*.pptx)", L"*.pptx"}, {L"Zip File(*.zip)",L"*.zip"},{ L"All File(*.*)",L"*.*" } });
		if (outNames.has_value() == false) {
			return false;
		}

		String outName = outNames.value();

		auto index = m_data->guis[2].radioButton(L"rbBmpMode").checkedItem;
		uint32 mode = index.value_or(0);
		auto minAlpha = m_data->guis[2].slider(SL_AMIN).valueInt;
		auto result = makeDots(fileName, outName, mode, minAlpha);

		return result;
	}


	/// <summary>`ȊO̍XVȂ</summary>
	void update() override {
			const GUI& gui0 = m_data->guis[0];
			const GUI& gui1 = m_data->guis[1];
			const GUI& gui2 = m_data->guis[2];

			gui0.text(L"tx1").text = L"*," + Pad(System::FrameCount() % 1000, {3,L'0'}) + L"f";
			gui0.text(L"txfps").text = Format(Profiler::FPS());

			double pt = 0.25 * gui1.slider(SL_THICKQ).valueInt;
			gui1.text(SL_THICKQ).text = Format(PyFmt, L"{0} pt", pt);

			if (gui1.button(L"btSet30").pushed) {
				gui1.slider(SL_FOVD).setValue(30);
			}
			gui1.text(SL_FOVD).text = Format(gui1.slider(SL_FOVD).valueInt);
			gui2.text(SL_AMIN).text = Format(gui2.slider(SL_AMIN).valueInt);

			if (Input::KeyF1.clicked) {
				m_data->curPYR.x = Math::HalfPi;
				m_data->curPYR.y = 0;
				m_data->basePYR = m_data->curPYR;
			}
			if (Input::KeyF2.clicked) {
				m_data->curPYR.x = 0;
				m_data->curPYR.y = 0;
				m_data->basePYR = m_data->curPYR;
			}
			if (Input::KeyF3.clicked) {
				m_data->curPYR.x = 0;
				m_data->curPYR.y = -Math::HalfPi;
				m_data->basePYR = m_data->curPYR;
			}


			if (s3d::Dragdrop::HasItems()) {
				auto files = s3d::Dragdrop::GetFilePaths();
				for (const auto fileName : files) {
					auto ext = FileSystem::Extension(fileName);
					if (ext == L"mqo") {
						m_data->colorsets.clear();
						auto result = loadLines(fileName, m_data->colorsets);
						String s = (result) ? L"ǂݍ݂܂" : L"ǂݍ݂Ɏs܂";
						MessageBox::Show(s);
					}
					else {
						auto result = saveDots(fileName);
						String s = (result) ? L"o͂܂" : L"o͂Ɏs܂";
						MessageBox::Show(s);
					}
				}
			}

			{ // Jړ
			  //// EhbOŃJ
				auto nowPos = Mouse::Pos();

				int mw = Mouse::Wheel();
				if (mw) {
					m_data->camDist += mw * 20.0;
					if (m_data->camDist <= 0.0) {
						m_data->camDist = 0.0;
					}
				}

				if (Input::MouseR.pressed) {

					if (m_data->prePos) { // ÂĉĂƂ
										  // 𔽉f
						auto prePos = m_data->prePos.value();
						m_data->curPYR.y = m_data->basePYR.y + (nowPos.x - prePos.x) * 0.01;
						m_data->curPYR.x = m_data->basePYR.x + (nowPos.y - prePos.y) * 0.01;
						m_data->curPYR.x = Clamp<double>(m_data->curPYR.x, -HalfPi, HalfPi);
					}
					else { // VƂ
						m_data->prePos.emplace(nowPos);
					}

				}
				else {
					if (m_data->prePos) { // Ƃ
						m_data->basePYR = m_data->curPYR;
						m_data->prePos.reset();
					}
					else { // NbNĂȂƂ
					}
				}
			}

			if (gui1.button(L"btSave").pushed) {
				Optional<Color> col;
				if (m_data->guis[1].checkBox(L"cbForceCol").checked(0)) {
					col = Palette::Black;
				}
				auto result = saveLines(m_data->guis[1].slider(SL_THICKQ).valueInt, col);
				String s = (result) ? L"o͂܂" : L"s܂";
				MessageBox::Show(s);
			}
	};
	void draw() const override {
		// J씽f
		double dist = m_data->camDist;
		Camera cam;
		cam.fovDegree = m_data->guis[1].slider(SL_FOVD).valueInt;
		cam.lookat = m_data->camTarget;
		cam.up = { 0, 1, 0 };
		Vec3 v = { 0, dist * -Sin(m_data->curPYR.x), dist * Cos(m_data->curPYR.x) };
		cam.pos = { v.z * Sin(m_data->curPYR.y), v.y, v.z * Cos(m_data->curPYR.y) };
		cam.pos = -cam.pos + cam.lookat;
		Graphics3D::SetCamera(cam);

		if (m_data->guis[0].checkBox(L"cbAxis").checked(0)) {
			double axis = 100;
			Line3D(Vec3(0, 0, 0), Vec3(axis, 0, 0)).drawForward(Palette::Red);
			Line3D(Vec3(0, 0, 0), Vec3(0, axis, 0)).drawForward(Palette::Green);
			Line3D(Vec3(0, 0, 0), Vec3(0, 0, axis)).drawForward(Palette::Blue);
		}

		for (const auto& colorLine : m_data->colorsets) {
			auto col = colorLine.col;
			if (m_data->guis[1].checkBox(L"cbForceCol").checked(0)) {
				col = Palette::Black;
			}
			for (const auto& line : colorLine.lines) {
				line.drawForward(col);
			}
		}

	};

	/// <summary>摜hbg</summary>
	/// <param name="fileName">摜t@C</param>
	/// <param name="mode">F̌ߕ</param>
	/// <param name="enableAlphaMin">L̃l̉D0ƕsx͋CɂȂ</param>
	bool makeDots(const String& fileName, const String& outName, int mode, uint32 enableAlphaMin) const {
		try {
			Image img(fileName);
			if (img.isEmpty) {
				return false;
			}

			int width = s3d::Min(MAXDOTSIDE, img.width);
			int height = s3d::Min(MAXDOTSIDE, img.height);
			int imgOffsetX = (img.width - width) / 2;
			int imgOffsetY = (img.height - height) / 2;

			int cellSize = (height <= 32) ? UNIT * 4 : UNIT * 2;

			int offsetX = UNIT * 8;
			int offsetY = UNIT * 8;

			Array<DotBox> dbs;


			// [hƂ̕ύX
			bool isColorExc = false;
			Color exc(255, 255, 255);

			switch (mode) {
			case 1: // 
				isColorExc = true;
				exc = img.getPixel(0, 0);
				break;
			case 2: // 
				isColorExc = true;
				break;
			case 3: // 
				isColorExc = true;
				exc = Color(0, 0, 0);
				break;
			}


			for (int y = 0; y < height; ++y) {
				for (int x = 0; x < width; ++x) {
					DotBox db;
					db.col = img.getPixel(y + imgOffsetY, x + imgOffsetX);

					if (db.col.a < enableAlphaMin) {
						continue;
					}
					if (isColorExc && db.col == exc) {
						continue;
					}

					db.rc.x = x * cellSize + offsetX;
					db.rc.y = y * cellSize + offsetY;
					db.rc.w = cellSize;
					db.rc.h = cellSize;
					dbs.push_back(db);
				}
			}

			Array<LineLine> lls;
			return savePptx(dbs, lls, outName, UNITPT);
		}
		catch (...) {
			Println(L"makeDots exception");
		}
		return false;
	}

	/// <summary>1s4  4x4 ̐</summary>
	Vec4 mulLH(const Vec4& v, const Mat4x4& m) const {
		Vec4 ret;
		ret.x = v.x * m.r[0].m128_f32[0] + v.y * m.r[1].m128_f32[0] + v.z * m.r[2].m128_f32[0] + v.w * m.r[3].m128_f32[0];
		ret.y = v.x * m.r[0].m128_f32[1] + v.y * m.r[1].m128_f32[1] + v.z * m.r[2].m128_f32[1] + v.w * m.r[3].m128_f32[1];
		ret.z = v.x * m.r[0].m128_f32[2] + v.y * m.r[1].m128_f32[2] + v.z * m.r[2].m128_f32[2] + v.w * m.r[3].m128_f32[2];
		ret.w = v.x * m.r[0].m128_f32[3] + v.y * m.r[1].m128_f32[3] + v.z * m.r[2].m128_f32[3] + v.w * m.r[3].m128_f32[3];
		return ret;
	};

	/// <summary>ݕێĂ郉C3Dt@Co</summary>
	/// <param name="thickQ">1/4 pt Pʂ̑</param>
	bool saveLines(int thickQ, Optional<Color>& force) const {

		auto outNames = Dialog::GetSave({ { L"PowerPoint(*.pptx)", L"*.pptx" },{ L"Zip File(*.zip)",L"*.zip" },{ L"All File(*.*)",L"*.*" } });
		if (outNames.has_value() == false) {
			return false;
		}
		String outName = outNames.value();

		try {
			Mat3x3<double> invH;
			Mat3x3<double> invP;
			invH.setHead(-m_data->curPYR.y);
			invP.setPitch(-m_data->curPYR.x);
			Mat3x3<double> cm = Mat3x3<double>::mulMat(invH, invP);

			Vec3 t(0.0, 0.0, m_data->camDist);

			Mat4x4 proj = Mat4x4::Identity();
			double ang = Math::Radians((double)m_data->guis[1].slider(SL_FOVD).valueInt);
			double near = 1.0;
			double far = 100000.0;
			double aspect = ((double)SL_W) / (double)SL_H;
			proj.r[0].m128_f32[0] = 1.0 / tan(ang * 0.5) / aspect;
			proj.r[1].m128_f32[1] = 1.0 / tan(ang * 0.5);
			proj.r[2].m128_f32[2] = far / (far - near);
			proj.r[2].m128_f32[3] = 1.0;
			proj.r[3].m128_f32[2] = -near * far / (far - near);
			proj.r[3].m128_f32[3] = 0.0;

			Array<LineLine> lls;

			for (const auto& set : m_data->colorsets) {
				auto col = force.value_or(set.col);
				for (const auto& line : set.lines) {
					auto b = Vec4(cm.leftMuled(line.begin) + t, 1);
					auto e = Vec4(cm.leftMuled(line.end) + t, 1);
					b = mulLH(b, proj);
					e = mulLH(e, proj);

					if (b.w != 0 && e.w != 0) {
						b *= 1.0 / b.w;
						e *= 1.0 / e.w;

						if (b.z >= 0.0 && e.z >= 0.0) {
							auto b2 = Vec2((b.x + 1.0) * SL_W * 0.5, (1.0 - b.y) * SL_H * 0.5);
							auto e2 = Vec2((e.x + 1.0) * SL_W * 0.5, (1.0 - e.y) * SL_H * 0.5);

							LineLine ll;
							ll.col = col;
							ll.line.begin = b2;
							ll.line.end = e2;
							ll.calcRect();
							lls.push_back(ll);
						}
					}
				}
			}

			Array<DotBox> dbs;
			savePptx(dbs, lls, outName, thickQ);
			return true;
		}
		catch (...) {
			Println(L"saveLines exception");
		}
		return false;
	}


	/// <summary>C𐶐ăt@Co</summary>
	/// <param name="fileName">.mqot@C</param>
	/// <param name="targets">󂯎z</param>
	bool loadLines(const String& fileName, Array<ColorLine3Ds>& targets) {
		try {
			MqoFiler filer;
			MqoFile mqo;
			TextReader tr(fileName);
			auto result = filer.loadMqo(tr, mqo);
			if (result == false) {
				Println(L"makeLines load error");
				return false;
			}

			Array<Array<Point>> idxss;
			int mtlNum = mqo.materials.size();
			for (int i = 0; i < mtlNum; ++i) {
				ColorLine3Ds target;
				target.col = mqo.materials[i].col;
				targets.push_back(target);

				Array<Point> idxs;
				idxss.push_back(idxs);
			}
			if (mtlNum <= 0) {
				mtlNum = 1;
				ColorLine3Ds target;
				target.col = Palette::Black;
				targets.push_back(target);

				Array<Point> idxs;
				idxss.push_back(idxs);
			}

			// CfbNXXg
				for (const auto& obj : mqo.objects) {

					// d`FbNp̃CfbNXi[zz
					Array<Array<Point>> idxss;
					for (int i = 0; i < mtlNum; ++i) {
						Array<Point> idxs;
						idxss.push_back(idxs);
					}

					for (const auto& face : obj.faces) {
						int mtlIdx = s3d::Clamp(face.materialIndex, 0, mtlNum - 1);
						auto n = face.indices.size();
						if (n == 2) {
							// ڌ^
							Point p2(face.indices[0], face.indices[1]);
							if (p2.x > p2.y) {
								auto tmp = p2.x;
								p2.x = p2.y;
								p2.y = tmp;
							}
							if (!AnyOf(idxss[mtlIdx], [&](const auto& a) {
								return (a.x == p2.x && a.y == p2.y);
							})) {
								idxss[mtlIdx].push_back(p2);
							}
						}
						else { // ʎw
							for (int i = 0; i < n; ++i) {
								Point p2(face.indices[i], face.indices[(i + 1) % n]);
								if (p2.x > p2.y) {
									auto tmp = p2.x;
									p2.x = p2.y;
									p2.y = tmp;
								}
									if (!AnyOf(idxss[mtlIdx], [&](const auto& a) {
										return (a.x == p2.x && a.y == p2.y);
									})) {
										idxss[mtlIdx].push_back(p2);
									}

							}
						}
					}

					// ǉ
					for (int i = 0; i < mtlNum; ++i) {
						for (int j = 0; j < idxss[i].size(); ++j) {
							const auto& pt = idxss[i][j];
							Line3D l;
							l.begin = zReved(obj.vertices[pt.x]);
							l.end = zReved(obj.vertices[pt.y]);
							targets[i].lines.push_back(l);
						}
					}

				}

				{
					auto dist = mqo.scene.pos.distanceFrom(mqo.scene.lookat);
					if (dist > 0.0) {
						m_data->camDist = dist;
					}
					m_data->camTarget = zReved(mqo.scene.lookat);

					//m_data->curPYR.x = -mqo.scene.pich;
					//m_data->curPYR.y = -mqo.scene.head;
					//m_data->basePYR = m_data->curPYR;
				}

			return true;
		}
		catch (...) {
			Println(L"loadLines exception");
		}
		return false;
	}

};

/// <summary>C֐</summary>
void Main() {
	MyApp myApp;
	myApp.add<MainScene>(L"MainScene");

	while (System::Update()) {
		if (!myApp.updateAndDraw()) {
			break;
		}
	}
}
