/**
 * @file mqoparser.hpp
 * @date 2017-05-28
 */

// Copyright (C) 2016-2017 Usagi
// Licensed under the MIT License.

// for Siv3D August 2016 v2

#pragma once

#include <Siv3D.hpp>

#define STRBUF (4096)
#define SIG L"Metasequoia Document"


/// <summary>wb_</summary>
struct MqoHeader {
	String sig = SIG;
	String type = L"Text";
	double ver = 1.1;
};

/// <summary>TlC</summary>
struct MqoThumbnail {
	int width = 128;
	int height = 128;
	int bitNum = 24;
	String colorFormat = L"rgb";
	String compress = L"raw";
	Array<String> lines;
};

/// <summary>Cg</summary>
struct MqoLight {
	Vec3 dir = Vec3(0.408, 0.408, 0.816);
	ColorF color = ColorF(1);
};

/// <summary>̃V[</summary>
struct MqoScene {
	Vec3 pos = Vec3(0, 0, 1500);
	Vec3 lookat = Vec3(0, 0, 0);
	double head = -0.5236;
	double pich = 0.5236;
	double bank = 0;
	int ortho = 0;
	double zoom2 = 5.0000;
	s3d::ColorF amb = ColorF(0.250, 0.250, 0.250);
	double frontclip = 225.00002;
	double backclip = 45000;
	Array<MqoLight> dirlights;
};

/// <summary>ގg</summary>
struct MqoMaterialEx {
	/// <summary>{ގ̃CfbNX</summary>
	int index;
	String shadertype = L"hlsl";
	String shadername = L"pmd";

	/// <summary>shaderparam</summary>
	Array<String> paramLines;

	/// <summary>ToonB1`10̒lB</summary>
	int toon = 1;
	/// <summary>GbWB01̒lB</summary>
	int edge = 1;
};

typedef MqoMaterialEx MqoMaterialEx2;

/// <summary>1̍ގs</summary>
struct MqoMaterial {
	/// <summary>ގ</summary>
	String name = L"mat";
	/// <summary>VF[_[ށBftHg3łȂ-1ɂB</summary>
	int shader = -1;
	/// <summary>ʂǂ</summary>
	int dbls = 0;
	/// <summary>F</summary>
	s3d::ColorF col = ColorF(1);
	/// <summary>gU̖邳</summary>
	double dif = 0.8000;
	/// <summary>̖邳</summary>
	double amb = 0.600;
	/// <summary>Ȕ̖邳</summary>
	double emi = 0.000;
	/// <summary>F̖邳</summary>
	double spc = 0.000;
	/// <summary>̋</summary>
	double power = 5.00;
	/// <summary>ʔ</summary>
	double reflect = 0.0;
	/// <summary>ܗ</summary>
	double refract = 0.0;
	/// <summary>͗l</summary>
	Optional<String> tex;
	/// <summary></summary>
	Optional<String> aplane;
	/// <summary>ʉ</summary>
	Optional<String> bump;

	/// <summary>F</summary>
	s3d::ColorF spc_col = ColorF(1);
	/// <summary>F</summary>
	s3d::ColorF amb_col = ColorF(1);

	/// <summary>ގ</summary>
	Optional<MqoMaterialEx2> ex2;
};


/// <summary>uu</summary>
struct MqoBlob {
	String type;
	double weight;
	Vec3 pos;
	Vec3 scale;
	Vec3 angle;
	double power = 1;
};

/// <summary>1̖</summary>
struct MqoFace {
	/// <summary>_CfbNX</summary>
	Array<int> indices;
	/// <summary>ގCfbNX</summary>
	int materialIndex = 0;

	/// <summary>UIDl</summary>
	int uid;

	/// <summary>_UV</summary>
	Array<Vec2> uvs;
	/// <summary>ȖʂȂǂ̃EFCg</summary>
	Array<double> crss;


	/// <summary>OpB[JCfbNXB0`(n-1)B⊮p</summary>
	Array<Array<int>> tris;
};

/// <summary>1̃IuWFNg</summary>
struct MqoObject {
	String name;
	/// <summary>̏ꍇ݂͑Ȃꍇ</summary>
	int uid = -1;
	int depth = 0;
	int folding = 0;
	Vec3 scale = Vec3(1, 1, 1);
	Vec3 rotation = Vec3(0, 0, 0);
	Vec3 translation = Vec3(0, 0, 0);
	int patch = 0;
	int patchtri;
	int segment;
	int visible = 15;
	int locking = 0;
	int shading = 1;
	double facet = 59.5;
	s3d::ColorF color = ColorF(0.898, 0.498, 0.698);
	int color_type = 0;
	int mirror = 0;
	int mirror_axis = 1;
	double mirror_dis = 100;
	Array<Vec3> vertices;
	Array<MqoFace> faces;

	/// <summary>ǉ</summary>
	Array<int> uids;
};


/// <summary>1̃f[^</summary>
struct MqoFile {
	/// <summary>[h̃tpX</summary>
	String path;

	/// <summary>wb_</summary>
	MqoHeader header;
	/// <summary>1̃TlC</summary>
	Optional<MqoThumbnail> thumbnail;
	/// <summary>1̃V[</summary>
	MqoScene scene;

	//MqoBackImage backImage;
	/// <summary>ގ</summary>
	Array<MqoMaterial> materials;
	/// <summary>ǂݎ莞̓ގ</summary>
	Array<MqoMaterialEx2> materialexs;
	Array<MqoObject> objects;
	/// <summary>^{[</summary>
	Array<MqoBlob> blobs;

	/// <summary>zȂǂNA</summary>
	void clear() {
		thumbnail.reset();
		materials.clear();
		materialexs.clear();
		objects.clear();
		blobs.clear();
	}
};

/// <summary>.mqot@Cǂݎ֐Q</summary>
struct MqoFiler {
	/// <summary>l3o</summary>
	Vec3 parseFloating(String& line) const {
		Vec3 ret;
		auto vs = line.trim().split(L' ');
		auto n = vs.size();
		if (n >= 1) {
			ret.x = FromString<double>(vs[0]);
		}
		if (n >= 2) {
			ret.y = FromString<double>(vs[1]);
		}
		if (n >= 3) {
			ret.z = FromString<double>(vs[2]);
		}
		return ret;
	}

	/// <summary>L[Đl3o</summary>
	Vec3 parseKeyFloating(String& line) const {
		Vec3 ret;
		auto vs = line.trim().split(L' ');
		auto n = vs.size();
		if (n >= 2) {
			ret.x = FromString<double>(vs[1]);
		}
		if (n >= 3) {
			ret.y = FromString<double>(vs[2]);
		}
		if (n >= 4) {
			ret.z = FromString<double>(vs[3]);
		}
		return ret;
	}

	/// <summary>_1sp[X</summary>
	Vec3 parseVertex(String& line) const {
		return parseFloating(line);
	}

	/// <summary>ʂ1sp[X</summary>
	MqoFace parseFace(String& line) const {
		MqoFace ret;

		std::wstring ws(line.trim().c_str());
		std::wsmatch wsm;

		int n = 0;
		std::wregex name(LR"(^(\d+))");
		if (std::regex_search(ws, wsm, name)) {
			if (wsm.size() >= 2) {
				n = FromString<int>(wsm[1].str());
			}
		}

		std::wregex v(LR"(V\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, v)) {
			if (wsm.size() >= 2) {
				String s(wsm[1].str());
				auto ss = s.split(L' ');
				for (auto elm : ss) {
					ret.indices.push_back(FromString<int>(elm));
				}
			}
		}

		std::wregex m(LR"(M\((\d+)\))");
		if (std::regex_search(ws, wsm, m)) {
			if (wsm.size() >= 2) {
				ret.materialIndex = FromString<int>(wsm[1].str());
			}
		}

		std::wregex uid(LR"(UID\((\d+)\))");
		if (std::regex_search(ws, wsm, uid)) {
			if (wsm.size() >= 2) {
				ret.materialIndex = FromString<int>(wsm[1].str());
			}
		}

		std::wregex uv(LR"(UV\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, uv)) {
			if (wsm.size() >= 2) {
				String s(wsm[1].str());
				auto ss = s.split(L' ');
				int num = ss.size();
				for (int i = 0; i < num / 2; ++i) {
					Vec2 v2(FromString<double>(ss[i * 2]), FromString<double>(ss[i * 2 + 1]));
					ret.uvs.push_back(v2);
				}
			}
		}

		/*
		std::wregex col(LR"(COL\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, col)) {
			if (wsm.size() >= 2) {
				//
			}
		}*/

		std::wregex crs(LR"(CRS\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, crs)) {
			if (wsm.size() >= 2) {
				String s(wsm[1].str());
				auto ss = s.split(L' ');
				for (auto elm : ss) {
					ret.crss.push_back(FromString<double>(elm));
				}
			}
		}

		return ret;
	}

	/// <summary>1s̍ގp[X</summary>
	MqoMaterial parseMaterial(String& line) const {
		MqoMaterial ret;

		std::wstring ws(line.trim().c_str());
		std::wsmatch wsm;

		std::wregex name(LR"(^\"([^\"]*)\")");
		if (std::regex_search(ws, wsm, name)) {
			if (wsm.size() >= 2) {
				ret.name = wsm[1].str();
			}
		}

		std::wregex shader(LR"(shader\((\d+)\))");
		if (std::regex_search(ws, wsm, shader)) {
			if (wsm.size() >= 2) {
				ret.shader = FromString<int>(wsm[1].str());
			}
		}

		std::wregex dbls(LR"(dbls\((\d+)\))");
		if (std::regex_search(ws, wsm, dbls)) {
			if (wsm.size() >= 2) {
				ret.dbls = FromString<int>(wsm[1].str());
			}
		}

		std::wregex col(LR"(col\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, col)) {
			if (wsm.size() >= 2) {
				String cols = wsm[1].str();
				auto ss = cols.split(L' ');
				if (ss.size() >= 4) {
					ret.col.r = FromString<double>(ss[0]);
					ret.col.g = FromString<double>(ss[1]);
					ret.col.b = FromString<double>(ss[2]);
					ret.col.a = FromString<double>(ss[3]);

					ret.spc_col.r = ret.col.r;
					ret.spc_col.g = ret.col.g;
					ret.spc_col.b = ret.col.b;

					ret.amb_col.r = ret.col.r;
					ret.amb_col.g = ret.col.g;
					ret.amb_col.b = ret.col.b;
				}
			}
		}

		std::wregex spc_col(LR"(spc_col\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, spc_col)) {
			if (wsm.size() >= 2) {
				String cols = wsm[1].str();
				auto ss = cols.split(L' ');
				if (ss.size() >= 3) {
					ret.spc_col.r = FromString<double>(ss[0]);
					ret.spc_col.g = FromString<double>(ss[1]);
					ret.spc_col.b = FromString<double>(ss[2]);
				}
			}
		}

		std::wregex amb_col(LR"(amb_col\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, amb_col)) {
			if (wsm.size() >= 2) {
				String cols = wsm[1].str();
				auto ss = cols.split(L' ');
				if (ss.size() >= 3) {
					ret.amb_col.r = FromString<double>(ss[0]);
					ret.amb_col.g = FromString<double>(ss[1]);
					ret.amb_col.b = FromString<double>(ss[2]);
				}
			}
		}

		std::wregex dif(LR"(dif\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, dif)) {
			if (wsm.size() >= 2) {
				ret.dif = FromString<double>(wsm[1].str());
			}
		}

		std::wregex amb(LR"(amb\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, amb)) {
			if (wsm.size() >= 2) {
				ret.amb = FromString<double>(wsm[1].str());
			}
		}

		std::wregex emi(LR"(emi\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, emi)) {
			if (wsm.size() >= 2) {
				ret.emi = FromString<double>(wsm[1].str());
			}
		}

		std::wregex spc(LR"(spc\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, spc)) {
			if (wsm.size() >= 2) {
				ret.spc = FromString<double>(wsm[1].str());
			}
		}

		std::wregex power(LR"(power\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, power)) {
			if (wsm.size() >= 2) {
				ret.power = FromString<double>(wsm[1].str());
			}
		}

		std::wregex reflect(LR"(reflect\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, reflect)) {
			if (wsm.size() >= 2) {
				ret.reflect = FromString<double>(wsm[1].str());
			}
		}

		std::wregex refract(LR"(refract\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, refract)) {
			if (wsm.size() >= 2) {
				ret.refract = FromString<double>(wsm[1].str());
			}
		}

		std::wregex tex(LR"(tex\(\"([^\"]*)\"\))");
		if (std::regex_search(ws, wsm, tex)) {
			if (wsm.size() >= 2) {
				ret.tex.emplace(wsm[1].str());
			}
		}

		std::wregex aplane(LR"(aplane\(\"([^\"]*)\"\))");
		if (std::regex_search(ws, wsm, aplane)) {
			if (wsm.size() >= 2) {
				ret.aplane.emplace(wsm[1].str());
			}
		}

		std::wregex bump(LR"(bump\(\"([^\"]*)\"\))");
		if (std::regex_search(ws, wsm, bump)) {
			if (wsm.size() >= 2) {
				ret.bump.emplace(wsm[1].str());
			}
		}

		return ret;
	}

	/// <summary>Scene`Nǂݎ</summary>
	/// <returns>isEof̏ԂԂ</returns>
	bool parseScene(TextReader& r,MqoFile& f) const {
		bool isEof = false;
		while (!isEof) {
			Optional<String> line = r.readLine();
			if (line.has_value() == false) {
				isEof = true;
				break;
			}

			String s = line.value().trim();
			if (s.includes(L"}")) {
				break;
			}

			auto ss = s.split(L' ');
			if (ss.size() < 2) {
				continue;
			}
			String key = ss[0];
			String v = ss[1];

			if (key == L"dirlights") {
				MqoLight light;
				while (!isEof) {
					line = r.readLine();
					if (line.has_value() == false) {
						isEof = true;
						break;
					}

					s = line.value().trim();
					if (s.includes(L"}")) {
						f.scene.dirlights.push_back(light);
						break;
					}

					if (s.includes(L"{")) {
						while (!isEof) {
							line = r.readLine();
							if (line.has_value() == false) {
								isEof = true;
								break;
							}

							s = line.value().trim();
							if (s.includes(L"}")) {
								break;
							}

							ss = s.split(L' ');
							key = ss[0];
							if (ss.size() >= 4) {
								Vec3 v3(Parse<double>(ss[1]),
									Parse<double>(ss[2]),
									Parse<double>(ss[3]));
								if (key == L"dir") {
									light.dir = v3;
								}
								else if (key == L"color") {
									light.color = ColorF(v3);
								}
							}

						}
					}

				} // dirlights

			}
			else if (key == L"head") {
				f.scene.head = Parse<double>(v);
			}
			else if (key == L"ortho") {
				f.scene.ortho = Parse<int>(v);
			}
			else if (key == L"zoom2") {
				f.scene.zoom2 = Parse<double>(v);
			}
			else if (key == L"amb") {
				if (ss.size() >= 4) {
					f.scene.amb = ColorF(Parse<double>(ss[1]),
						Parse<double>(ss[2]),
						Parse<double>(ss[3]));
				}
			}
		}

		return isEof;
	}

	/// <summary>Object`Nǂݎ</summary>
	/// <returns>isEof̏ԂԂ</returns>
	bool parseObject(TextReader& r,MqoObject& obj) const {
		bool isEof = false;

		while (!isEof) {
			auto line = r.readLine();
			if (line.has_value() == false) {
				isEof = true;
				break;
			}

			String s = line.value().trim();
			if (s.includes(L"}")) {
				break;
			}

			if (s.includes(L"BVertex ")) {
				isEof = true;
				break;
			}

			auto ss = s.split(L' ');
			int num = ss.size();
			String key = ss[0];

			if (key == L"uid") {
				if (num >= 2) {
					obj.uid = Parse<int>(ss[1]);
				}
			}
			else if (key == L"depth") {
				if (num >= 2) {
					obj.depth = Parse<int>(ss[1]);
				}
			}
			else if (key == L"folding") {
				if (num >= 2) {
					obj.folding = Parse<int>(ss[1]);
				}
			}
			else if (key == L"scale") {
				if (num >= 4) {
					obj.scale = parseKeyFloating(s);
				}
			}
			else if (key == L"rotation") {
				if (num >= 4) {
					obj.rotation = parseKeyFloating(s);
				}
			}
			else if (key == L"translation") {
				if (num >= 4) {
					obj.translation = parseKeyFloating(s);
				}
			}
			else if (key == L"visible") {
				if (num >= 2) {
					obj.visible = Parse<int>(ss[1]);
				}
			}
			else if (key == L"locking") {
				if (num >= 2) {
					obj.locking = Parse<int>(ss[1]);
				}
			}
			else if (key == L"shading") {
				if (num >= 2) {
					obj.shading = Parse<int>(ss[1]);
				}
			}
			else if (key == L"facet") {
				if (num >= 2) {
					obj.facet = Parse<double>(ss[1]);
				}
			}
			else if (key == L"color") {
				if (num >= 4) {
					obj.color = ColorF(parseKeyFloating(s));
				}
			}
			else if (key == L"color_type") {
				if (num >= 2) {
					obj.color_type = Parse<int>(ss[1]);
				}
			}
			else if (key == L"vertex") {
				while (!isEof) {
					line = r.readLine();
					if (line.has_value() == false) {
						isEof = true;
						break;
					}
					s = line.value().trim();
					if (s.includes(L"}")) {
						break;
					}

					auto v = this->parseVertex(s);
					obj.vertices.push_back(v);
				}
			}
			else if (key == L"vertexattr") {
				isEof = this->parseVertexAttr(r, obj);

				/*
				while (!isEof) {
					line = r.readLine();
					if (!line) {
						isEof = true;
						break;
					}
					s = line.value().trim();
					if (s.includes(L"}")) {
						break;
					}


				}
				*/
			}
			else if (key == L"face") {
				while (!isEof) {
					line = r.readLine();
					if (line.has_value() == false) {
						isEof = true;
						break;
					}
					s = line.value().trim();
					if (s.includes(L"}")) {
						break;
					}

					auto face = this->parseFace(s);
					obj.faces.push_back(face);
				}
			}
			else if (s.includes(L"{")) {
				isEof = skipUnknown(r);
			}

		}
		return isEof;
	}

	/// <summary>MaterialEx2`Nǂݎ</summary>
	/// <returns>isEof̏ԂԂ</returns>
	bool parseMaterialEx2s(TextReader& r, MqoFile& f) const {
		bool isEof = false;

		while (!isEof) {
			auto line = r.readLine();
			if (line.has_value() == false) {
				isEof = true;
				break;
			}

			String s = line.value().trim();
			if (s.startsWith(L"}")) {
				break;
			}
			else if (s.startsWith(L"material")) {
				auto ss = s.split(L' ');
				if (ss.size() >= 2) {
					MqoMaterialEx2 m;
					m.index = FromString<int>(ss[1]);
					while (!isEof) {
						line = r.readLine();
						if (line.has_value() == false) {
							isEof = true;
							break;
						}
						s = line.value().trim();
						if (s.startsWith(L"}")) {
							break;
						} else if (s.startsWith(L"shadertype")) {
							ss = s.split(L'\"');
							if (ss.size() >= 3) {
								m.shadertype = ss[1];
							}
						}
						else if (s.startsWith(L"shadername")) {
							ss = s.split(L'\"');
							if (ss.size() >= 3) {
								m.shadername = ss[1];
							}
						}
						else if (s.startsWith(L"shaderparam")) {
							while (!isEof) {
								Optional<String> lineSP = r.readLine();
								if (lineSP.has_value() == false) {
									isEof = true;
									break;
								}
								s = lineSP.value().trim();
								if (s.startsWith(L"}")) {
									break;
								} else if (s.startsWith(L"Toon")) {
									ss = s.split(L' ');
									if (ss.size() >= 3) {
										m.toon = FromString<int>(ss[2]);
									}
								}
								else if (s.startsWith(L"Edge")) {
									ss = s.split(L' ');
									if (ss.size() >= 3) {
										m.edge = FromString<int>(ss[2]);
									}
								}

							}
						}
						else if (s.startsWith(L"shadermappingparam")) {
							isEof = skipUnknown(r);
						}

					}
					f.materialexs.push_back(m);
				}
			}

		}
		return isEof;
	}


	/// <summary>_ǂݔ΂</summary>
	bool parseVertexAttr(TextReader& r, MqoObject& obj) const {
		bool isEof = false;
		int layer = 1;
		while (!isEof) {
			auto line = r.readLine();
			if (line.has_value() == false) {
				isEof = true;
				break;
			}
			auto s = line.value();
			if (s.includes(L"{")) {
				++layer;
			}
			else if (s.includes(L"}")) {
				--layer;
				if (layer <= 0) {
					break;
				}
			}
		}
		return isEof;
	}

	/// <summary>mȂ`Nǂݔ΂</summary>
	/// <returns>isEof̏ԂԂ</returns>
	bool skipUnknown(TextReader& r) const {
		bool isEof = false;
		int layer = 1;
		while (!isEof) {
			auto line = r.readLine();
			if (line.has_value() == false) {
				isEof = true;
				break;
			}
			auto s = line.value();
			if (s.includes(L"{")) {
				++layer;
			}
			else if (s.includes(L"}")) {
				--layer;
				if (layer <= 0) {
					break;
				}
			}
		}
		return isEof;
	}

	/// <summary>.mqot@Cǂݎ</summary>
	/// <param name="r">Shift_JIS̃eLXg[_</param>
	/// <param name="f">ʂmۂ</param>
	bool loadMqo(TextReader& r,MqoFile& f) const {
		try {
			Optional<String> line;

			bool isEof = false;

			line = r.readLine();
			if (line.has_value() == false) {
				return false;
			}
			if (line.value() != SIG) {
				return false;
			}

			line = r.readLine();

			while (!isEof) {
				line = r.readLine();
				if (line.has_value() == false) {
					isEof = true;
					break;
				}
				String s = line.value();
				if (s.startsWith(L"Eof")) {
					isEof = true;
					break;
				}

				if (s.startsWith(L"Thumbnail")) {
					MqoThumbnail tn;
					while (!isEof) {
						line = r.readLine();
						if (line.has_value() == false) {
							isEof = true;
							break;
						}

						s = line.value();
						if (s.includes(L"}")) {
							f.thumbnail = tn;
							break;
						}

						tn.lines.push_back(s);
					}
				}
				else if (s.startsWith(L"Scene")) {
					isEof = parseScene(r,f);
				}
				else if (s.startsWith(L"BackImage")) {
					isEof = skipUnknown(r);
				}
				else if (s.startsWith(L"Material ")) {
					while (!isEof) {
						line = r.readLine();
						if (line.has_value() == false) {
							isEof = true;
							break;
						}

						s = line.value().trim();
						if (s.includes(L"}")) {
							break;
						}

						auto m = this->parseMaterial(s);
						f.materials.push_back(m);
					}
				}
				else if (s.startsWith(L"MaterialEx2") || s.startsWith(L"MaterialEx")) {
					isEof = parseMaterialEx2s(r, f);
				}
				else if (s.startsWith(L"Object")) {
					MqoObject obj;
					auto ss = s.split(L'\"');
					if (ss.size() >= 3) {
						obj.name = ss[1];
					}

					isEof = parseObject(r, obj);
					if (isEof == false) {
						f.objects.push_back(obj);
					}
				}
				else if (s.startsWith(L"Blob")) {
					isEof = skipUnknown(r);
				}
				else if (s.includes(L"{")) {
					isEof = skipUnknown(r);
				}
			}

			// ގ𔽉f
			int matNum = f.materials.size();
			for (auto& m : f.materialexs) {
				int index = m.index;
				if (0 <= index && index < matNum) {
					f.materials[index].ex2 = m;
				}
			}

		}
		catch (...) {
			return false;
		}
		return true;
	}

};
